/* scanner for a c2txt translator */

%{
/* need this for the call to atof() below */
#include "c2txt.h"
#include "c2txt.tab.h"
#include <math.h>
#include <string.h>
#define YY_NO_UNPUT
int yyparse();
int define_flag=FALSE;
%}

DIGIT    [0-9]
ID       [a-zA-Z_][a-zA-Z0-9_]*

%x comment
%%

"/*"   {BEGIN(comment); return COMMENT_BEGIN;}
<comment>[^*]* {CSTR; return COMMENT;}
<comment>"*"+[^*/]* {CSTR; return COMMENT;}
<comment>"*"+"/"        {BEGIN(INITIAL); CSTR; yylval[strlen(yylval)-2]=0; return COMMENT_END;}

"UWORD_32bits" {RSTR(UWORD_32BITS_STR); return TYPE;}
"unsigned long" {RSTR(UNSIGNED_LONG_STR); return TYPE;}
"unsigned int" {RSTR(UNSIGNED_INT); return TYPE;}
"UWORD_16bits" {RSTR(UWORD_16BITS); return TYPE;}
"unsigned short" {RSTR(UNSIGNED_SHORT); return TYPE;}
"short" {RSTR(SHORT); return TYPE;}
"UBYTE_08bits" {RSTR(UBYTE_08BITS); return TYPE;}
"unsigned char" {RSTR(UNSIGNED_CHAR); return TYPE;}
"char" {RSTR(CHAR); return TYPE;}
"union aword" {RSTR(UNION_AWORD); return TYPE;}
"void" {RSTR(VOID); return TYPE;}
"inline" {return INLINE;}
"static" {return STATIC;}

"=============== bf_tab.h ==============" {CSTR;return SPECIAL;}
"************** TEST VECTORS ***********************************" {CSTR;return SPECIAL;}
"This is a test vector."|"Plaintext is"|"The key is" {CSTR;return SPECIAL;}
"This is another test vector." {CSTR;return SPECIAL;}


#define {define_flag=TRUE; return DEFINE;}
#ifdef {return IFDEF;}
#endif {return ENDIF;}
struct {return STRUCT;}
#include {return INCLUDE;}
for {return FOR;}
return {return RETURN;}

0x[0-9a-fA-F]+ {CSTR; return HEX;}
0x[0-9a-fA-F]+l {CSTR; return BIG_HEX;}

-*{DIGIT}+    {CSTR; return NUM;}
{ID} {CSTR; return NAME;}

"++" {return PLUS_PLUS;}
"+=" {return AUGMENTED_ADD;}
"^=" {return AUGMENTED_XOR;}

[+;\[\](){};:*,#\.^=<%&] {return yytext[0];}

\"[^\"]*\" {CSTR; return STRING;}
\<[^\<]*\> {CSTR; return STRING;}

\n {if(define_flag) {define_flag=FALSE; return DEFINE_END;}}

[ \t]+ /*{CSTR; return WHITE_SPACE;}*/

.           printf( "Unrecognized character: %s\n", yytext );

%%

int main(int argc, char **argv)
{
  ++argv, --argc;  /* skip over program name */
  if ( argc > 0 )
    yyin = fopen( argv[0], "r" );
  else
    yyin = stdin;

  init();

  yyparse();
  return 0;
}
