/* grammar for a c2txt translator */

%{
#include "c2txt.h"
#include <string.h>
#include <stdlib.h>
int yyerror(char *str);
int yylex();
%}

/* BISON Declarations */
%token COMMENT_BEGIN COMMENT COMMENT_END
%token DEFINE IFDEF ENDIF TYPE INCLUDE DEFINE_END
%token AUGMENTED_XOR '+' '^' ';' '[' ']' '(' ')' PLUS_PLUS AUGMENTED_ADD
%token '{' '}' ';' ':' '*' ',' '<' '>' '.' '#' '=' '<' '%' '&'
%token NAME POINTER_NAME
%token NUM WHITE_SPACE STRING HEX BIG_HEX
%token STRUCT INLINE FOR RETURN STATIC
%token SPECIAL
%left AUGMENTED_XOR '=' AUGMENTED_ADD
%left '<'
%left '+' '^'
%left '%'
/* Grammar follows */
%%
input:    /* empty string */
     | input statement_nl {printf("%s",$2);}
;

compound: '{' statements '}' {
  ASTR4($$,COMPOUND_START,$2,COMPOUND_END,D0("{compound (..)}"));
}
;

statements: statement_nl
          | statement_nl statements {ASTR2($$,$1,$2);}
;

statement_nl: statement {ASTR2($$,$1,"\n\n");}
;

statement:  comment
          | define_nl
          | ifdef
          | endif
          | include
          | union_definition
          | variable_definition
          | struct_definition
          | function_definition
          | expr ';' {ASTR3($$,$1,EXPR_END,D0("{expr ;}"));}
          | INLINE {ASTR1($$,INLINE_START);}
          | for
          | return
          | special
;

num: NUM {ASTR1($$,$1);}
   | HEX {ASTR2($$,HEX_STR_START,hex2nstr($1));}
   | BIG_HEX {ASTR2($$,BIG_HEX_STR_START,hex2nstr($1));}
;

expr: num /*{ASTR3($$,EXPR_NUM_START,$1,EXPR_NUM_END);}*/
    | variable {$$=$1;}
    | '(' expr ')' {
      ASTR4($$,EXPR_PARENTHESIS_START,$2,EXPR_PARENTHESIS_END,
            D0("{expr (..)}"));
    }
    | expr '%' expr {
      ASTR4($$,$1,EXPR_REMAINDER,D0("{expr ..%%%%..}"),$3);
    }
    | expr '+' expr {
      ASTR4($$,$1,EXPR_ADD,D0("{expr ..+..}"),$3);
    }
    | expr '^' expr {
      ASTR4($$,$1,EXPR_XOR,D0("{expr ..^..}"),$3);
    }
    | function_call
    | variable AUGMENTED_XOR expr {
      ASTR4($$,$1,EXPR_AUGMENTED_XOR,D0("{expr ..^=..}"),$3);
    }
    | variable '=' expr {
      ASTR4($$,$1,EXPR_SET,D0("{expr ..=..}"),$3);
    }
    | variable AUGMENTED_ADD expr {
      ASTR4($$,$1,EXPR_AUGMENTED_ADD,D0("{expr ..+=..}"),$3);
    }
    | expr '<' expr {
      ASTR4($$,$1,EXPR_LESS_THAN,D0("{expr ..<..}"),$3);
    }
    | PLUS_PLUS variable {
      ASTR3($$,EXPR_PLUS_PLUS,$2,D0("{expr ++..}"));
    }
;

variable: NAME {
          ASTR6($$,VARIABLE_NAME_START,get_type($1),
                VARIABLE_NAME_MIDDLE,get_name($1),
                VARIABLE_NAME_END,
                D1("{variable %s}",$1));}
        | NAME '[' expr ']' {
          ASTR8($$,VARIABLE_TABLE_START,$3,
                VARIABLE_TABLE_MIDDLE1,get_name($1),
                VARIABLE_TABLE_MIDDLE2,get_type($1),
                VARIABLE_TABLE_END,
                D1("{variable %s[..]}",$1));
        }
        | NAME '[' expr ']' '[' expr ']' {
          ASTR10($$,VARIABLE_TABLE2_START,$3,
                 VARIABLE_TABLE2_MIDDLE1,get_name($1),
                 VARIABLE_TABLE2_MIDDLE2,$6,
                 VARIABLE_TABLE2_MIDDLE3,get_type($1),
                 VARIABLE_TABLE_END,
                 D1("{variable %s[..][..]}",$1));
        }
        | variable '.' NAME {
          ASTR7($$,$1,VARIABLE_PART_MIDDLE1,get_name($3),
                VARIABLE_PART_MIDDLE2,get_type($3),
                VARIABLE_PART_END,
                D1("{variable .. . %s}",$3));
        }
        | variable '#' '#' NAME {
          ASTR5($$,$1,VARIABLE_NAME_ADD_MIDDLE,get_name($4),
                VARIABLE_NAME_ADD_END,
                D1("{variable ..##%s}",$4));
        }
        | '*' NAME {
          ASTR6($$,VARIABLE_POINTER_START,get_type($2),
                VARIABLE_POINTER_MIDDLE,get_name($2),
                VARBIALE_POINTER_END,
                D1("{variable *%s}",$2));
        }
        | '&' NAME {
          ASTR4($$,VARIABLE_REFERENCE_START,get_name($2),
                VARBIALE_REFERENCE_END,
                D1("{variable &%s}",$2));
        }
;

comment: COMMENT_BEGIN comment_string COMMENT_END {
  ASTR4($$,COMMENT_BEGIN_STR,$2,$3,COMMENT_END_STR);
}
;

comment_string: COMMENT
              | COMMENT comment_string {ASTR2($$,$1,$2);}
;

define_nl: define DEFINE_END
         | define comment DEFINE_END {ASTR2($$,$1,$2);}
;

define: define_name_num
      | define_name_string
      | define_type
      | define_macro
;

        /*#define MAXKEYBYTES 56 -> There is 'James' that has '56' leafs.*/
define_name_num: DEFINE NAME num {
  ASTR8($$,DEFINE_NAME_NUM_START,add_id($2,DEFINE_NAME_NUM_TYPE),
        DEFINE_NAME_NUM_MIDDLE1,$2,
        DEFINE_NAME_NUM_MIDDLE2,$3,
        DEFINE_NAME_NUM_END,
        D1("{define_name_num %s ..}",$2));
}
;

define_name_string: DEFINE NAME STRING {
  ASTR8($$,DEFINE_NAME_STRING_START,add_id($2,DEFINE_NAME_STRING_TYPE),
        DEFINE_NAME_STRING_MIDDLE1,$2,
        DEFINE_NAME_STRING_MIDDLE2,$3,
        DEFINE_NAME_STRING_END,
        D2("{define_name_string %s %s}",$2,$3));
}
;

define_type: DEFINE TYPE TYPE {
  ASTR6($$,DEFINE_TYPE_START,$2,DEFINE_TYPE_MIDDLE,$3,DEFINE_TYPE_END,
        D2("{define_type %s %s}",$2,$3));
}
;

macro_par: NAME {
  ASTR4($$,add_id($1,DEFINE_NAME_MACRO_ARG_TYPE),
        MACRO_PAR_START,$1,
        D1("{macro_par %s}",$1));}
;

macro_pars: macro_par
          | macro_par ',' macro_pars {
  ASTR4($$,$1,MACRO_PARS_MIDDLE,D0("{macro_pars .. , .. }"),$3);
}
;

define_macro: DEFINE NAME '(' macro_pars ')' expr {
  ASTR8($$,DEFINE_MACRO_START,$2,
        DEFINE_MACRO_MIDDLE,$4,
        DEFINE_MACRO_END1,$6,
        DEFINE_MACRO_END2,
        D1("{define_macro %s (..)}",$2));
}
;

ifdef: IFDEF NAME {ASTR4($$,IFDEF_START,$2,IFDEF_END,D1("{ifdef %s}",$2));}
;

endif: ENDIF {ASTR2($$,ENDIF_START,D0("{endif}"));}
;

include: INCLUDE STRING {
  ASTR4($$,INCLUDE_START,$2,INCLUDE_END,D1("{include %s}",$2));
}
;

argument: expr {ASTR2($$,$1,D0("{argument}"));}
;

arguments: argument
         | argument ',' arguments {
  ASTR4($$,$1,ARGUMENTS_MIDDLE,D0("{arguments .. , ..}"),$3);
}
;

function_call: NAME '(' arguments ')' {
  ASTR6($$,FUNCTION_CALL_START,$1,
        FUNCTION_CALL_MIDDLE,$3,
        FUNCTION_CALL_END,
        D1("{function_call %s (..)}",$1));
}
;

union_definition: TYPE compound ';' {
  ASTR6($$,DEFINE_UNION_AWORD_START,$1,DEFINE_UNION_AWORD_MIDDLE,
        $2,DEFINE_UNION_AWORD_END,D1("{union_definition %s ..;}",$1));
}
;

variable_definition: TYPE NAME variable_init {
                     ASTR9($$,VARIABLE_DEFINITION_ORDINARY_START,add_id($2,$1),
                           VARIABLE_DEFINITION_ORDINARY_MIDDLE,$2,
                           VARIABLE_DEFINITION_ORDINARY_MIDDLE2,$1,
                           VARIABLE_DEFINITION_ORDINARY_END,$3,
                           D2("{variable_definition %s %s ..}",$1,$2));
                   }
                   | TYPE NAME '[' ']' variable_init {
                     ASTR9($$,VARIABLE_DEFINITION_ARRAY_START,add_id($2,$1),
                           VARIABLE_DEFINITION_ARRAY_MIDDLE,$2,
                           VARIABLE_DEFINITION_ARRAY_MIDDLE2,$1,
                           VARIABLE_DEFINITION_ARRAY_END,$5,
                           D2("{variable_definition %s %s[] ..}",$1,$2));
                   }
                   | TYPE NAME '[' expr ']' variable_init {
                     ASTR11($$,VARIABLE_DEFINITION_ARRAYE_START,add_id($2,$1),
                            VARIABLE_DEFINITION_ARRAYE_MIDDLE,$2,
                            VARIABLE_DEFINITION_ARRAYE_MIDDLE2,$4,
                            VARIABLE_DEFINITION_ARRAYE_MIDDLE3,$1,
                            VARIABLE_DEFINITION_ARRAYE_END,$6,
                            D2("{variable_definition %s %s[..] ..}",$1,$2));
                   }
                   | TYPE NAME '[' expr ']' '[' expr ']' variable_init {
                     ASTR13($$,VARIABLE_DEFINITION_ARRAY2E_START,add_id($2,$1),
                            VARIABLE_DEFINITION_ARRAY2E_MIDDLE1,$2,
                            VARIABLE_DEFINITION_ARRAY2E_MIDDLE2,$4,
                            VARIABLE_DEFINITION_ARRAY2E_MIDDLE3,$7,
                            VARIABLE_DEFINITION_ARRAY2E_MIDDLE4,$1,
                            VARIABLE_DEFINITION_ARRAY2E_END,$9,
                            D2("{variable_definition %s %s[..] [..] ..}",$1,$2));
                   }
                   | TYPE NAME ':' num variable_init {
                     ASTR11($$,VARIABLE_DEFINITION_BITS_START,add_id($2,$1),
                            VARIABLE_DEFINITION_BITS_MIDDLE,$2,
                            VARIABLE_DEFINITION_BITS_MIDDLE2,$1,
                            VARIABLE_DEFINITION_BITS_MIDDLE3,$4,
                            VARIABLE_DEFINITION_BITS_END,$5,
                            D2("{variable_definition %s %s:.. ..}",$1,$2));
                   }
                   | STATIC variable_definition {
                     ASTR3($$,VARIABLE_DEFINITION_STATIC_START,$2,
                           D0("{variable_definition static ..}"));
                   }
;

vlist: expr ',' {ASTR3($$,$1,VLIST_END,D0("{vlist .. ,}"));}
     | expr ',' vlist {ASTR4($$,$1,VLIST_MIDDLE,D0("{vlist .. , ..}"),$3);}
;

variable_init: ';' {ASTR2($$,VARIABLE_INIT_END,D0("{variable_init ;}"));}
             | '=' '{' vlist '}' ';' {
               ASTR4($$,VARIABLE_INIT_VLIST_START,$3,
                     VARIABLE_INIT_VLIST_END,
                     D0("{variable_init =(..)}"));
             }
             | '=' STRING ';' {
               ASTR4($$,VARIABLE_INIT_STRING_START,$3,
                     VARIABLE_INIT_STRING_END,
                     D0("{variable_init =..}"));
             }
;

parameters: parameter
          | parameter ',' parameters {
  ASTR4($$,$1,PARAMETERS_MIDDLE,D0("{parameters .. , ..}"),$3);
}
;

parameter: TYPE {ASTR2($$, $1,D1("{parameter %s}",$1));}
        | TYPE NAME {
          ASTR6($$,add_id($2,$1),
                PARAMETER_NORMAL_MIDDLE1,$2,
                PARAMETER_NORMAL_MIDDLE2,$1,
                D2("{parameter %s %s}",$1,$2));}
        | TYPE '*' NAME {
          ASTR7($$,PARAMETER_POINTER_START,add_id($3,$1),
                PARAMETER_POINTER_MIDDLE1,$3,
                PARAMETER_POINTER_MIDDLE2,$1,
                D2("{parameter %s *%s}",$1,$3));
        }
        | TYPE NAME '[' ']' {
          ASTR7($$,add_id($2,$1),
                PARAMETER_ARRAY_MIDDLE1,$2,
                PARAMETER_ARRAY_MIDDLE2,$1,
                PARAMETER_ARRAY_END,
                D2("{parameter %s %s[]}",$1,$2));
        }
;

struct_definition: STRUCT compound NAME ';' {
  ASTR10($$,DEFINE_STRUCT_START,DEFINE_STRUCT_TYPE,
         DEFINE_STRUCT_MIDDLE,$2,
         DEFINE_STRUCT_END1,add_id($3,DEFINE_STRUCT_TYPE),
         DEFINE_STRUCT_END2,$3,
         DEFINE_STRUCT_END3,
         D1("{struct_definition (..) %s ;}",$3));
}
;

function_definition: TYPE NAME '(' parameters ')' ';' {
                     ASTR8($$,FUNCTION_DEFINITION_NOBODY_START,$2,
                           FUNCTION_DEFINITION_NOBODY_MIDDLE,$4,
                           FUNCTION_DEFINITION_NOBODY_END1,$1,
                           FUNCTION_DEFINITION_NOBODY_END2,
                           D2("{function_definition %s %s (..);}",$1,$2));
                   }
                   | TYPE NAME '(' parameters ')' compound {
                     ASTR10($$,FUNCTION_DEFINITION_BODY_START,$2,
                            FUNCTION_DEFINITION_BODY_MIDDLE1,$4,
                            FUNCTION_DEFINITION_BODY_MIDDLE2,$6,
                            FUNCTION_DEFINITION_BODY_END1,$1,
                            FUNCTION_DEFINITION_BODY_END2,
                            D2("{function_definition %s %s (..) (..)}",$1,$2));
                   }
;

for: FOR '(' expr ';' expr ';' expr ')' compound {
  ASTR9($$,FOR_START,$3,
        FOR_MIDDLE1,$5,
        FOR_MIDDLE2,$7,
        FOR_END,$9,
        D0("{for(..;..;..) (..)}"));
}
;

return: RETURN expr ';' {ASTR4($$,RETURN_START,$2,RETURN_END,D0("{return ..;}"));}
;

special: SPECIAL {
         ASTR4($$,SPECIAL_START,$1,
               SPECIAL_END,
               D1("{special %s}",$1));}
       | SPECIAL STRING {
         ASTR6($$,SPECIAL_START,$1,
               SPECIAL_MIDDLE,$2,
               SPECIAL_STRING_END,
               D2("{special %s %s}",$1,$2));
       }
       | SPECIAL STRING '.' {
         ASTR6($$,SPECIAL_START,$1,
               SPECIAL_MIDDLE,$2,
               SPECIAL_POINT,
               D2("{special %s %s.}",$1,$2));
       }
;


%%

void init()
{
  read_names("names.lst");
  add_id("bf_P","melons"); /* blowfish.c first uses these before defining them ;-) */
  add_id("bf_S","melons"); /* wanted it to be able to process it unchanged */
#if DEBUG
  debStr=malloc(DEB_STR_SIZE);
  if(!debStr) {
    fprintf(stderr,"Not enough memory\n");
    exit(-1);
  }
#endif
  /*yydebug=1;*/
}
