/* grammar for a txt2c translator -*-C-*- */

%{
#include "txt2c.h"
#include <string.h>
#include <stdlib.h>
int yyerror(char *str);
int yylex();
#define TYPE_NOTSUPPLIED "type not supplied"
int vlist_count=0;
%}

/* BISON Declarations */
%token CHARACTER



define(`def',`%token $1')
define(`def2',`')

include(`x2x.m4')

/* ' comment for C major mode*/

%left EXPR_AUGMENTED_XOR EXPR_SET EXPR_AUGMENTED_ADD
%left EXPR_LESS_THAN
%left EXPR_ADD EXPR_XOR
%left EXPR_REMAINDER

/* Grammar follows */
%%
input:    /* empty string */
     | input statement_nl {printf("%s",$2);}
;

compound: COMPOUND_START statements COMPOUND_END {
  ASTR3($$,"{\n",$2,"}");
}
;

statements: statement_nl
          | statement_nl statements {ASTR2($$,$1,$2);}
;

statement_nl: statement {ASTR2($$,$1,"\n");}
;

statement:  comment
          | define
          | ifdef
          | endif
          | include
          | union_definition
          | variable_definition
          | struct_definition
          | function_definition
          | expr EXPR_END {ASTR2($$,$1,";");}
          | INLINE_START {ASTR1($$,"inline");}
          | for
          | return
          | special
;

character: CHARACTER
         | END_SENTENCE {ASTR1($$,".");} /* needed for points */
         | AND {ASTR1($$," and ");} /* ugly hack: needed for this: /* TODO: test with a through z as key and plain text */

string: character {ASTR1($$,$1);/*printf("(C1:%s)\n",$1);*/}
      | character string {ASTR2($$,$1,$2);/*printf("(C2:%s,%s)\n",$1,$2);*/}
;

type: UWORD_32BITS_STR {ASTR1($$,"UWORD_32bits");}
    | UNSIGNED_LONG_STR {ASTR1($$,"unsigned long");}
    | UNSIGNED_INT {ASTR1($$,"unsigned int");}
    | UWORD_16BITS {ASTR1($$,"UWORD_16bits");}
    | UNSIGNED_SHORT {ASTR1($$,"unsigned short");}
    | SHORT {ASTR1($$,"short");}
    | UBYTE_08BITS {ASTR1($$,"UBYTE_08bits");}
    | UNSIGNED_CHAR {ASTR1($$,"unsigned char");}
    | CHAR {ASTR1($$,"char");}
    | UNION_AWORD {ASTR1($$,"union aword");}
    | VOID {ASTR1($$,"void");}
    | DEFINE_NAME_NUM_TYPE {ASTR1($$,TYPE_NOTSUPPLIED);}
    | DEFINE_NAME_STRING_TYPE {ASTR1($$,TYPE_NOTSUPPLIED);}
    | DEFINE_NAME_MACRO_ARG_TYPE {ASTR1($$,TYPE_NOTSUPPLIED);}
    | DEFINE_STRUCT_TYPE {ASTR1($$,TYPE_NOTSUPPLIED);}
;

num: string /*NUM*/ {ASTR1($$,$1);}
   | HEX_STR_START string /*NUM*/ {ASTR1($$,nstr2hexstr($2));}
   | BIG_HEX_STR_START string /*NUM*/ {ASTR2($$,nstr2hexstr($2),"l");}
;

expr: num {ASTR1($$,$1);}
    | variable {$$=$1;}
    | EXPR_PARENTHESIS_START expr EXPR_PARENTHESIS_END {
      ASTR3($$,"(",$2,")");
    }
    | expr EXPR_REMAINDER expr {
      ASTR3($$,$1,"%",$3);
    }
    | expr EXPR_ADD expr {
      ASTR3($$,$1,"+",$3);
    }
    | expr EXPR_XOR expr {
      ASTR3($$,$1,"^",$3);
    }
    | function_call
    | variable EXPR_AUGMENTED_XOR expr {
      ASTR3($$,$1,"^=",$3);
    }
    | variable EXPR_SET expr {
      ASTR3($$,$1,"=",$3);
    }
    | variable EXPR_AUGMENTED_ADD expr {
      ASTR3($$,$1,"+=",$3);
    }
    | expr EXPR_LESS_THAN expr {
      ASTR3($$,$1,"<",$3);
    }
    | EXPR_PLUS_PLUS variable {
      ASTR2($$,"++",$2);
    }
;

variable: VARIABLE_NAME_START type VARIABLE_NAME_MIDDLE string /*NAME*/ VARIABLE_NAME_END {
            ASTR1($$,get_id($4));
          }
        | VARIABLE_TABLE_START expr VARIABLE_TABLE_MIDDLE1 string /*NAME*/ VARIABLE_TABLE_MIDDLE2 type END_SENTENCE {
          ASTR4($$,get_id($4),"[",$2,"]");
        }
        | VARIABLE_TABLE2_START expr VARIABLE_TABLE2_MIDDLE1 string /*NAME*/ VARIABLE_TABLE2_MIDDLE2 expr VARIABLE_TABLE2_MIDDLE3 type END_SENTENCE {
          ASTR6($$,get_id($4),"[",$2,"][",$6,"]");
        }
        | variable VARIABLE_PART_MIDDLE1 string /*NAME*/ VARIABLE_PART_MIDDLE2 type VARIABLE_PART_END {
          ASTR3($$,$1,".",get_id($3));
        }
        | variable VARIABLE_NAME_ADD_MIDDLE string /*NAME*/ VARIABLE_NAME_ADD_END {
          ASTR3($$,$1,"##",get_id($3));
        }
        | VARIABLE_POINTER_START type VARIABLE_POINTER_MIDDLE string /*NAME*/ VARBIALE_POINTER_END {
          ASTR2($$,"*",get_id($4));
        }
        | VARIABLE_REFERENCE_START string /*NAME*/ VARBIALE_REFERENCE_END {
          ASTR2($$,"&",get_id($2));
        }
;

comment: COMMENT_BEGIN_STR string QUOTE_END {
  ASTR3($$,"/*",$2,"*/");
}
;

define: define_name_num
      | define_name_string
      | define_type
      | define_macro
;

define_name_num: THERE_IS string CODENAMED string DEFINE_NAME_NUM_MIDDLE2 num DEFINE_NAME_NUM_END 
{
  C($2,add_id($4,TYPE_NOTSUPPLIED));
  ASTR4($$,"#define ",$4," ",$6);
}
;

define_name_string: DEFINE_NAME_STRING_START string CODENAMED string DEFINE_NAME_STRING_MIDDLE2 string DEFINE_NAME_STRING_END {
  C($2,add_id($4,TYPE_NOTSUPPLIED));
  ASTR4($$,"#define ",$4," ",$6);
}
;

define_type: DEFINE_TYPE_START type DEFINE_TYPE_MIDDLE type END_SENTENCE {
  ASTR4($$,"#define ",$2," ",$4);
}
;

macro_par: string CODENAMED string {
  C($1,add_id($3,TYPE_NOTSUPPLIED));
  ASTR1($$,$3);
}
;

macro_pars: macro_par
          | macro_par MACRO_PARS_MIDDLE macro_pars {
  ASTR3($$,$1,",",$3);
}
;

define_macro: DEFINE_MACRO_START string DEFINE_MACRO_MIDDLE macro_pars DEFINE_MACRO_END1 expr DEFINE_MACRO_END2 {
  ASTR6($$,"#define ",$2,"(",$4,") ",$6);
}
;

ifdef: IFDEF_START string IFDEF_END {ASTR2($$,"#ifdef ",$2);}
;

endif: ENDIF_START {ASTR1($$,"#endif");}
;

include: INCLUDE_START string INCLUDE_END {
  ASTR2($$,"#include ",$2);
}
;

argument: expr
;

arguments: argument
         | argument ARGUMENTS_MIDDLE arguments {
  ASTR3($$,$1,",",$3);
}
;

function_call: FUNCTION_CALL_START string /*NAME*/ FUNCTION_CALL_MIDDLE arguments FUNCTION_CALL_END {
  ASTR4($$,$2,"(",$4,")");
}
;

union_definition: DEFINE_UNION_AWORD_START type DEFINE_UNION_AWORD_MIDDLE compound DEFINE_UNION_AWORD_END {
  ASTR4($$,$2," ",$4,";");
}
;

variable_definition: THERE_IS string CODENAMED string VARIABLE_DEFINITION_ORDINARY_MIDDLE2 type END_SENTENCE variable_init 
                   {
                     C($2,add_id($4,$6));
                     ASTR4($$,$6," ",$4,$8);
                   }
                   | THERE_IS string CODENAMED string VARIABLE_DEFINITION_ARRAY_MIDDLE2 type END_SENTENCE variable_init {
                     C($2,add_id($4,$6));
                     ASTR5($$,$6," ",$4,"[]",$8);
                   }
                   | THERE_IS string CODENAMED string VARIABLE_DEFINITION_ARRAYE_MIDDLE2 expr VARIABLE_DEFINITION_ARRAYE_MIDDLE3 type END_SENTENCE variable_init {
                     C($2,add_id($4,$8));
                     ASTR7($$,$8," ",$4,"[",$6,"]",$10);
                   }
                   | THERE_IS string CODENAMED string /*NAME*/ VARIABLE_DEFINITION_ARRAY2E_MIDDLE2 expr VARIABLE_DEFINITION_ARRAY2E_MIDDLE3 expr VARIABLE_DEFINITION_ARRAY2E_MIDDLE4 type END_SENTENCE variable_init {
                     C($2,add_id($4,$10));
                     ASTR9($$,$10," ",$4,"[",$6,"][",$8,"]",$12);
                   }
                   | THERE_IS string CODENAMED string VARIABLE_DEFINITION_BITS_MIDDLE2 type VARIABLE_DEFINITION_BITS_MIDDLE3 num VARIABLE_DEFINITION_BITS_END variable_init {
                     C($2,add_id($4,$6));
                     ASTR6($$,$6," ",$4,":",$8,$10);
                   }
                   | VARIABLE_DEFINITION_STATIC_START variable_definition {
                     ASTR2($$,"static ",$2);
                   }
;

vlist: expr VLIST_END {ASTR2($$,$1,",\n");vlist_count=0;}
     | expr VLIST_MIDDLE vlist {
       vlist_count++;
       if(vlist_count%4==0) {ASTR3($$,$1,",\n",$3);}
       else {ASTR3($$,$1,", ",$3);}
     }
;

variable_init: VARIABLE_INIT_END {ASTR1($$,";");}
             | '=' '{' vlist '}' ';' 
             | VARIABLE_INIT_VLIST_START vlist VARIABLE_INIT_VLIST_END {
               ASTR3($$,"={\n",$2,"};");
             }
             | VARIABLE_INIT_STRING_START string VARIABLE_INIT_STRING_END {
               ASTR3($$,"=",$2,";");
             }
;

parameters: parameter
          | parameter AND parameters {
  ASTR3($$,$1,", ",$3);
}
;

parameter: type
         | string CODENAMED string LIKING type {
           C($1,add_id($3,$5));
           ASTR3($$,$5," ",$3);
         }
         | PARAMETER_POINTER_START string CODENAMED string LIKING type {
           C($2,add_id($4,$6));
           ASTR3($$,$6," *",$4);
         }
         | string CODENAMED string LIKING type PARAMETER_ARRAY_END {
           C($1,add_id($3,$5));
           ASTR4($$,$5," ",$3,"[]");
         }
;

struct_definition: DEFINE_STRUCT_START DEFINE_STRUCT_TYPE DEFINE_STRUCT_MIDDLE compound DEFINE_STRUCT_END1 string CODENAMED string DEFINE_STRUCT_END3 
{
  C($6,add_id($8,TYPE_NOTSUPPLIED));
  ASTR4($$,"struct ",$4,$8,";");
}
;

function_definition: FUNCTION_DEFINITION_NOBODY_START string FUNCTION_DEFINITION_NOBODY_MIDDLE parameters FUNCTION_DEFINITION_NOBODY_END1 type FUNCTION_DEFINITION_NOBODY_END2 {
                     ASTR6($$,$6," ",$2,"(",$4,");");
                   }
                   | FUNCTION_DEFINITION_BODY_START string /*NAME*/ FUNCTION_DEFINITION_BODY_MIDDLE1 parameters FUNCTION_DEFINITION_BODY_MIDDLE2 compound FUNCTION_DEFINITION_BODY_END1 type FUNCTION_DEFINITION_BODY_END2 {
                     ASTR7($$,$8," ",$2,"(",$4,")\n",$6);
                   }
;

for: FOR_START expr FOR_MIDDLE1 expr FOR_MIDDLE2 expr FOR_END compound {
  ASTR8($$,"for(",$2,";",$4,";",$6,") ",$8);
}
;

return: RETURN_START expr RETURN_END {
  ASTR3($$,"return ",$2,";");
}
;

special: SPECIAL_START string QUOTE_END {ASTR1($$,$2);}
       | SPECIAL_START string SPECIAL_MIDDLE string SPECIAL_STRING_END {
         ASTR3($$,$2," ",$4);
       }
       | SPECIAL_START string SPECIAL_MIDDLE string SPECIAL_POINT {
         ASTR4($$,$2," ",$4,".");
       }
;

%%

void init()
{
  read_names("names.lst");
  add_id("bf_P","melons"); /* blowfish.c first uses these before defining them ;-) */
  add_id("bf_S","melons"); /* wanted it to be able to process it unchanged */
#if DEBUG
  debStr=malloc(DEB_STR_SIZE);
  if(!debStr) {
    fprintf(stderr,"Not enough memory\n");
    exit(-1);
  }
#endif
  /*yydebug=1;*/
}
